##############################################################################
# Imports
##############################################################################

import argparse
import subprocess
import termcolor

from . import ansible_common

##############################################################################
# Methods
##############################################################################


def parse_ros_args(args):
    """
    Parses args and launches playbooks depending on the arg parsing functionality.
    """
    termcolor.cprint("**********************************************************", "green")
    print("       {0}{1}{2}".format(termcolor.colored("This is the ", "green"),
                                    termcolor.colored("'pc-ros'", "white", attrs=["bold"]),
                                    termcolor.colored(" ansible playbook", "green")))
    termcolor.cprint("**********************************************************\n", "green")
    cmd = "ansible-playbook pc-ros.yml -K -i localhost, -c local"
    if args.verbose:
        cmd += " -vvv"
    print("{0}: {1}\n".format(termcolor.colored("Command", "cyan"),
                              termcolor.colored(cmd, "yellow")))
    subprocess.call(cmd, cwd=args.home, shell=True)
    print("Done")
    # print("  x: {0}".format(args.x))


def add_subparser(subparsers):
    """
    Add our own argparser to the parent.

    :param subparsers: the subparsers factory from the parent argparser.
    """
    ros_parser = subparsers.add_parser("pc-ros",
                                       description="Install, configure or update an existing ros distro.",  # this shows in the help for this command
                                       help="manage a ros installation",  # this shows in the parent parser
                                       formatter_class=argparse.ArgumentDefaultsHelpFormatter
                                       )
    ansible_common.add_ansible_arguments(ros_parser)
    ros_parser.set_defaults(func=parse_ros_args)

#     parser = argparse.ArgumentParser(description='Executor for ansible playbooks on a pc.',
#                                      epilog=show_epilog(),
#                                      formatter_class=argparse.ArgumentDefaultsHelpFormatter)  # RawTextHelpFormatter for whitespace preservation
#     parser.add_argument('-j', '--jobs', type=int, metavar='JOBS', default=None, nargs='?', help='Specifies the number of jobs (commands) to run simultaneously. Defaults to the environment variable ROS_PARALLEL_JOBS and falls back to the number of CPU cores.')
#     parser.add_argument('--force-cmake', action='store_true', help='Invoke "cmake" even if it has been executed before [false]')
#     parser.add_argument('-p', '--pre-clean', action='store_true', help='Clean build temporaries before making [false]')
#     parser.add_argument('-c', '--cmake-only', action='store_true', help='Do not compile, just force a re-run of cmake [false]')
#     group = parser.add_mutually_exclusive_group()
#     group.add_argument('-i', '--install', action='store_true', help='Run install step after making [false]')
#     group.add_argument('--track', choices=settings.VALID_TRACKS, dest='default_underlay', action='store', default=None, help='convenience equivalent for the --default-underlay option')
#     group.add_argument('--install-rosdeps-track', choices=settings.VALID_TRACKS, action='store', default=None, help='Install all rosdeps for the workspace sources and given track [None]')
#     group.add_argument('--install-rosdeps', action='store_true', help='Install all rosdeps for the workspace sources and track set by `yujin_tools_settings --get-default-track` [false]')
#     group.add_argument('-t', '--tests', action='store_true', help='Make tests [false]')
#     group.add_argument('-r', '--run-tests', action='store_true', help='Run tests (does not build them) [false]')
#     parser.add_argument('--strip', action='store_true', help='Strips binaries, only valid with --install')
#     parser.add_argument('--no-color', action='store_true', help='Disables colored ouput')
#     parser.add_argument('--target', default=None, help='Build against a particular target only')
#     parser.add_argument('--pkg', help='Invoke "make" on a specific package only')
#     parser.add_argument('--cmake-args', dest='cmake_args', nargs='*', type=str,
#         help='Arbitrary arguments which are passes to CMake. It must be passed after other arguments since it collects all following options.')
#     parser.add_argument('--make-args', dest='make_args', nargs='*', type=str,
#         help='Arbitrary arguments which are passes to make. It must be passed after other arguments since it collects all following options. This is only necessary in combination with --cmake-args since else all unknown arguments are passed to make anyway.')



##############################################################################
# Imports
##############################################################################

import os

##############################################################################
# Methods
##############################################################################


def ansible_playbooks_home():
    """
    Root directory of where the ansible playbooks package may be found. If a
    dev has cloned the package elsewhere, there are other handles for overriding
    this value.

    :returns: path to the playbooks package
    """
    return os.path.join(os.path.abspath(os.sep), "opt", "yujin", "ansible_playbooks")


def add_ansible_arguments(parser):
    """
    Any arguments for the parser that should be available for every single
    ansible call.

    :param argparse.ArgumentParser parser:
    """
    parser.add_argument('-v', '--verbose', action='store_true', help='execute the playbook with extra verbosity')
    parser.add_argument('--home', action='store', default=ansible_playbooks_home(), help='path to the ansible playbooks home')




from . import ansible
from . import ansible_common
from . import ansible_pc

from .init_workspace import init_workspace
from .init_build import init_build
from .make import make_main
from .make_isolated import make_isolated_main

# should be before imports, but current pep8 barfs at this
__version__ = '0.4.27'



##############################################################################
# Imports
##############################################################################

import os.path

import stat  # file permissions
import argparse
from argparse import RawTextHelpFormatter
import shutil

##############################################################################
# Local imports
##############################################################################

import console
import common
import settings

##############################################################################
# Methods
##############################################################################


def help_string():
    overview = 'This is a convenience script for auto-generating a catkin parallel build directory.\n\n'
    instructions = " \
 'yujin_init_build .' prepares conventional ros build directories in ./ linked to sources in ./src \n \
 'yujin_init_build .' prepares conventional ros build directories in ./ linked to sources in ./src \n \
 'yujin_init_build --release release' prepares a release directory in ./release linked to sources in ./src \n \
 'yujin_init_build debug ~/ecl/src' prepares a build directory in ./debug linked to sources in ~/ecl/src \n \
 'yujin_init_build --no-default-underlay --toolchain=arm-pc-linux-gnueabi arm' prepares a build directory in ./arm with the specified toolchain module \n \
 'yujin_init_build --underlays=~/cslam/src/native;~/ecl/src/native .' prepares a build directory in ./arm with the specified toolchain module \n \
 \n \
 Underlays automatically add /opt/ros/`yujin_tools_settings --get-default-track` unless '--no-default-underlay' is specified (in which case you should\n \
 ensure that the catkin sources are in your workspace or one of your underlays).\n \
"
    return overview + instructions


def parse_arguments():
    parser = argparse.ArgumentParser(description=help_string(), formatter_class=RawTextHelpFormatter)
    parser.add_argument('dir', nargs='?', default=".", help='directory to use for the parallel development space [./]')
    parser.add_argument('sources', nargs='?', default="src", help='directory where the sources reside [./src]')
    parser.add_argument('-r', '--release', action='store_true', help='build in Release mode instead of RelWithDebInfo [false]')
    parser.add_argument('-i', '--install', action='store', default='/not_set_directory', help='installation location [workspace/install]')
    parser.add_argument('-d', '--doc', action='store', default='/not_set_directory', help='document location [workspace/doc]')
    #  even though we don't set a default value here, it later gets set as the default underlay if not present and --no-default-underlay is not true
    parser.add_argument('-u', '--underlays', action='store', default='', help='semi-colon list of catkin workspaces to utilise, priority given from front to back')
    default_underlay_group = parser.add_mutually_exclusive_group()
    default_underlay_group.add_argument('-du', '--default-underlay', choices=settings.VALID_TRACKS, action='store', default=None, help='default the underlays to this track if catkin is not found [`yujin_tools_settings --get-default-track`]')
    default_underlay_group.add_argument('--track', choices=settings.VALID_TRACKS, dest='default_underlay', action='store', default=None, help='convenience equivalent for the --default-underlay option')
    default_underlay_group.add_argument('-n', '--no-default-underlay', action='store_true', help='do not automatically underlay with the default track setting [false]')
    parser.add_argument('-t', '--toolchain', action='store', default='', help='toolchain cmake module to load []')
    parser.add_argument('-p', '--platform', action='store', default='default', help='platform cmake cache module to load [default]')
    parser.add_argument('-c', '--clean', action='store_true', help='cleans the current or specified build directory [false]')
    parser.add_argument('--list-toolchains', action='store_true', help='list all currently available toolchain modules [false]')
    parser.add_argument('--list-platforms', action='store_true', help='list all currently available platform modules [false]')
    args = parser.parse_args()
    return args


def which(program):
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, unused_fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


def read_template(tmplf):
    f = open(tmplf, 'r')
    try:
        t = f.read()
    finally:
        f.close()
    return t


def fill_in_template(template, name, cwd):
    return template % locals()


def instantiate_template(filename, name, cwd):
    template_dir = os.path.join(os.path.dirname(__file__), 'templates', 'init_build')
    tmpl = read_template(os.path.join(template_dir, filename))
    contents = fill_in_template(tmpl, name, cwd)
    try:
        f = open(os.path.join(cwd, filename), 'w')
        f.write(contents.encode('utf-8'))
    finally:
        os.fchmod(f.fileno(), stat.S_IRWXU | stat.S_IRWXG)
        f.close()


def fill_in_config_cmake(template, config_build_type, config_devel, config_install_prefix, config_doc_prefix, config_underlays, config_override_file):
    return template % locals()


def instantiate_config_cmake(platform_content, build_path, config_build_type, config_install_prefix, config_doc_prefix, config_underlays):
    '''
      Copy the cache configuration template to the build path.

      @param platform_content : content to prepend to the default configuration file
      @param build_path : location of the build directory to save the config.cmake file.
    '''
    template_dir = os.path.join(os.path.dirname(__file__), 'cmake')
    template = read_template(os.path.join(template_dir, "config.cmake"))
    config_devel = os.path.join(build_path, 'devel')
    contents = fill_in_config_cmake(template, config_build_type, config_devel, config_install_prefix, config_doc_prefix, config_underlays, common.override_filename())
    config_cmake_file = os.path.join(build_path, "config.cmake")
    try:
        f = open(config_cmake_file, 'w')
        f.write(platform_content.encode('utf-8'))
        f.write(contents.encode('utf-8'))
    finally:
        f.close()


def print_build_details(build_dir, source_dir, install_prefix, doc_prefix, build_type, underlays, name, toolchain, platform):
    console.pretty_println("\n************************** Parallel Buildspace Details ***************************", console.bold)
    console.pretty_print(" -- Build directory : ", console.cyan)
    console.pretty_println(build_dir, console.yellow)
    console.pretty_print(" -- Source directory: ", console.cyan)
    console.pretty_println(source_dir, console.yellow)
    console.pretty_print(" -- Install prefix  : ", console.cyan)
    console.pretty_println(install_prefix, console.yellow)
    console.pretty_print(" -- Document prefix : ", console.cyan)
    console.pretty_println(doc_prefix, console.yellow)
    console.pretty_print(" -- Build Type      : ", console.cyan)
    console.pretty_println(build_type, console.yellow)
    underlays_list = underlays.split(';')
    console.pretty_print(" -- Underlays       : ", console.cyan)
    if len(underlays_list) == 0:
        console.pretty_println("-", console.yellow)
    else:
        underlay = underlays_list.pop(0)
        console.pretty_println(underlay, console.yellow)
        for underlay in underlays_list:
            console.pretty_print("                    : ", console.cyan)
            console.pretty_println(underlay, console.yellow)
    console.pretty_print(" -- Eclipse Name    : ", console.cyan)
    console.pretty_println(name, console.yellow)
    if not toolchain == "":
        console.pretty_print(" -- Toolchain       : ", console.cyan)
        console.pretty_println(toolchain, console.yellow)
    if not platform == "":
        console.pretty_print(" -- Platform        : ", console.cyan)
        console.pretty_println(platform, console.yellow)
    console.pretty_println("**********************************************************************************\n", console.bold)


def write_yujin_init_build_configuration(build_source_dir, source_dir):
    try:
        f = open(os.path.join(build_source_dir, '.yujin_init_build'), 'w')
        rel_path = os.path.relpath(source_dir, build_source_dir)
        f.write(rel_path.encode('utf-8'))
    finally:
        f.close()


def get_toolchains_or_platforms(base_path):
    '''
      Does a look up in the path for either toolchain or platform files.
      @param base_path : start of the search (ends in either toolchains or platforms).
    '''
    d = {}
    for (dirpath, unused_dirname, filenames) in os.walk(base_path):
        if dirpath != base_path:  # skip the root, only find files in subfolders
            family = os.path.basename(dirpath)
            d[family] = []
            for filename in filenames:
                d[family].append(os.path.splitext(filename)[0])  # leave off the .cmake extension
    return d


def list_toolchains():
    console.pretty_println("\n******************************** Toolchain List **********************************", console.bold)
    toolchains = get_toolchains_or_platforms(os.path.join(os.path.dirname(__file__), 'toolchains'))
    console.pretty_println("Official:", console.bold)
    for family in toolchains:
        for platform in toolchains[family]:
            console.pretty_print(" -- %s/" % family, console.cyan)
            console.pretty_println("%s" % platform, console.yellow)
    console.pretty_println("Custom:", console.bold)
    toolchains = get_toolchains_or_platforms(os.path.join(settings.yujin_tools_home(), 'toolchains'))
    if toolchains:
        for family in toolchains:
            for platform in toolchains[family]:
                console.pretty_print(" -- %s/" % family, console.cyan)
                console.pretty_println("%s" % platform, console.yellow)
    else:
        console.pretty_println(" -- ", console.cyan)
    console.pretty_println("**********************************************************************************\n", console.bold)


def list_platforms():
    console.pretty_println("\n********************************* Platform List **********************************", console.bold)
    platforms = get_toolchains_or_platforms(os.path.join(os.path.dirname(__file__), 'platforms'))
    console.pretty_println("Official:", console.bold)
    for family in platforms:
        for platform in platforms[family]:
            console.pretty_print(" -- %s/" % family, console.cyan)
            console.pretty_println("%s" % platform, console.yellow)
    console.pretty_println("Custom:", console.bold)
    platforms = get_toolchains_or_platforms(os.path.join(settings.yujin_tools_home(), 'platforms'))
    if platforms:
        for family in platforms:
            for platform in platforms[family]:
                console.pretty_print(" -- %s/" % family, console.cyan)
                console.pretty_println("%s" % platform, console.yellow)
    else:
        console.pretty_println(" -- ", console.cyan)
    console.pretty_println("**********************************************************************************\n", console.bold)


def init_configured_build(default_underlay, build_dir_="./", source_dir_="./src", underlays_="/opt/ros/groovy", install_prefix_="./install", doc_prefix_="./doc", release_=False, toolchain_="", platform_=""):
    '''
      This one is used with pre-configured parameters. Note that
      init_build generates parameters parsed from the command line and then
      calls this function.
    '''
    # Help us build the name for the eclipse workspace...usually we call in the workspace itself.
    workspace_dir = os.getcwd()

    ##########################
    # Build directory
    ##########################
    if os.path.isabs(build_dir_):
        build_dir = build_dir_
    else:
        if build_dir_ == "." or build_dir_ == "./":
            build_dir = os.getcwd()
        else:
            build_dir = os.path.join(os.getcwd(), build_dir_)
    if not os.path.isfile(os.path.join(build_dir, "Makefile")):
        if not os.path.isdir(build_dir):  # remember ./ is a valid build dir, even if it's not populated yet
            os.mkdir(build_dir)
    else:
        raise RuntimeError("This build directory is already initialised")

    ##########################
    # Source directory
    ##########################
    source_dir = os.path.abspath(source_dir_)
    build_source_dir = os.path.join(build_dir, 'src')
    if not os.path.isdir(source_dir):
        raise RuntimeError("Specified source space does not exist [" + source_dir + "]")
    if not os.path.isfile(os.path.join(source_dir, ".rosinstall")):
        raise RuntimeError("Could not find a valid source folder (must contain a .rosinstall file therein)'")
    if os.path.exists(build_source_dir):
        if not source_dir == build_source_dir:
            raise RuntimeError("The build directory already has a ./src directory which doesn't match the desired source directory [%s]" % source_dir)
    else:
        os.mkdir(build_source_dir)
        source_subdirectories = os.walk(source_dir).next()[1]
        for d in source_subdirectories:
            common.create_symlink(os.path.join(source_dir, d), os.path.join(build_source_dir, d))
    write_yujin_init_build_configuration(build_source_dir, source_dir)

    ##########################
    # Underlays
    ##########################
    try:
        env_underlays = os.environ['CMAKE_PREFIX_PATH']
    except KeyError:
        env_underlays = ""
    underlays_list = [underlay for underlay in underlays_.split(';') if underlay]
    env_underlays_list = [underlay for underlay in env_underlays.split(':') if underlay]
    for underlay in env_underlays_list:
        if underlay not in underlays_list:
            underlays_list.append(underlay)

    ##########################
    # Locate Catkin
    ##########################
    catkin_toplevel = None
    if os.path.isfile(os.path.join(source_dir, 'catkin', 'cmake', 'toplevel.cmake')):
        catkin_toplevel = os.path.join(source_dir, 'catkin', 'cmake', 'toplevel.cmake')
    else:
        catkin_toplevel, unused_catkin_python_path, unused_catkin_cmake_path = common.find_catkin(build_dir, underlays_list)

    ##########################
    # Add toplevel if exists
    ##########################
    if not catkin_toplevel:
        # Add the default underlay
        if default_underlay is not None and os.path.isfile(os.path.join("/opt/ros/%s" % default_underlay, 'share', 'catkin', 'cmake', 'toplevel.cmake')):
            catkin_toplevel = os.path.join("/opt/ros/%s" % default_underlay, 'share', 'catkin', 'cmake', 'toplevel.cmake')
            unused_catkin_python_path = os.path.join("/opt/ros/%s" % default_underlay, 'lib', 'python2.7', 'dist-packages')
            console.pretty_println("No catkin found, adding the default underlay (use yujin_tools_settings to change) [/opt/ros/%s]" % default_underlay, console.cyan)
            underlays_list.append("/opt/ros/%s" % default_underlay)
        else:
            raise RuntimeError("Could not find an underlying catkin installation.")
    common.create_symlink(catkin_toplevel, os.path.join(build_source_dir, "CMakeLists.txt"))
    underlays = ';'.join(underlays_list)

    ##########################
    # Other Args
    ##########################
    if install_prefix_ == "/not_set_directory":
        install_prefix = os.path.join(build_dir, "install")
    else:
        install_prefix = install_prefix_

    if doc_prefix_ == "/not_set_directory":
        doc_prefix = os.path.join(build_dir, "doc")
    else:
        doc_prefix = doc_prefix_

    if release_:
        build_type = "Release"
    else:
        build_type = "RelWithDebInfo"
    name = os.path.basename(workspace_dir) + "_" + os.path.basename(build_dir)

    print_build_details(build_dir, source_dir, install_prefix, doc_prefix, build_type, underlays, name, toolchain_, platform_)
    os.chdir(build_dir)

    ##########################
    # Toolchain
    ##########################
    if not toolchain_ == "":
        toolchains_dir = os.path.join(os.path.dirname(__file__), 'toolchains')
        custom_toolchains_dir = os.path.join(settings.yujin_tools_home(), 'toolchains')
        tmp_list = toolchain_.split('/')
        if len(tmp_list) != 2:
            raise RuntimeError("Toolchain specification invalid, must be <family>/<tuple> [%s]" % toolchain_)
        family = tmp_list[0]
        toolchain_tuple = tmp_list[1]
        toolchains = get_toolchains_or_platforms(toolchains_dir)
        custom_toolchains = get_toolchains_or_platforms(custom_toolchains_dir)
        if not family in toolchains and not family in custom_toolchains:
            raise RuntimeError("No toolchains available for family %s" % family)
        if family in custom_toolchains and toolchain_tuple in custom_toolchains[family]:
            toolchain_file = os.path.join(custom_toolchains_dir, family, toolchain_tuple + ".cmake")
        elif family in toolchains and toolchain_tuple in toolchains[family]:
            toolchain_file = os.path.join(toolchains_dir, family, toolchain_tuple + ".cmake")
        else:
            raise RuntimeError("Platform %s for family %s not available." % (family, toolchain_tuple))
        if os.path.isfile(toolchain_file):
            shutil.copy(toolchain_file, os.path.join(build_dir, "toolchain.cmake"))
        else:
            raise RuntimeError("Toolchain module not available [%s]" % toolchain_file)

    ##########################
    # Platform
    ##########################
    platform_content = ""
    platforms_dir = os.path.join(os.path.dirname(__file__), 'platforms')
    custom_platforms_dir = os.path.join(settings.yujin_tools_home(), 'platforms')
    if not platform_ == "default":
        tmp_list = platform_.split('/')
        if len(tmp_list) != 2:
            raise RuntimeError("Platform specification invalid, must be <family>/<platform type> [%s]" % platform_)
        family = tmp_list[0]
        platform = tmp_list[1]
        platforms = get_toolchains_or_platforms(platforms_dir)
        custom_platforms = get_toolchains_or_platforms(custom_platforms_dir)
        if not family in platforms and not family in custom_platforms:
            raise RuntimeError("No platforms available for family %s" % family)
        if family in custom_platforms and platform in custom_platforms[family]:
            platform_file = os.path.join(custom_platforms_dir, family, platform + ".cmake")
        elif family in platforms and platform in platforms[family]:
            platform_file = os.path.join(platforms_dir, family, platform + ".cmake")
        else:
            raise RuntimeError("Platform %s for family %s not available." % (family, platform))
    else:
        platform_file = os.path.join(platforms_dir, 'default.cmake')

    if os.path.isfile(platform_file):
        f = open(platform_file, 'r')
        try:
            platform_content = f.read()
        finally:
            f.close()
    else:
        raise RuntimeError("Platform configuration not available [%s]" % platform_)

    ##########################
    # Cache
    ##########################
    instantiate_config_cmake(platform_content, build_dir, build_type, install_prefix, doc_prefix, underlays)

    ##########################
    # Templates
    ##########################
    instantiate_template('.bashrc', name, build_dir)
    instantiate_template('konsole', name, build_dir)
    instantiate_template('gnome-terminal', name, build_dir)
    instantiate_template('eclipse', name, build_dir)
    instantiate_template('android-studio', name, build_dir)


def clean(dir_to_be_cleaned, dir_sources):
    if common.is_same_dir(dir_to_be_cleaned, os.getcwd()):
        if not os.path.isfile(os.path.join(dir_to_be_cleaned, 'config.cmake')):
            console.logerror("Could not clean the current directory [build artifacts do not exist]")
            return
        console.pretty_print("\nCleaning current directory of yujin_init_build artifacts : ", console.cyan)
        for f in [os.path.join(dir_to_be_cleaned, x) for x in ['config.cmake', 'eclipse', 'gnome-terminal', 'konsole', 'toolchain.cmake', '.bashrc', 'android-studio']]:
            if os.path.isfile(f):
                os.remove(f)
        for f in [os.path.join(dir_to_be_cleaned, dir_sources, x) for x in ['CMakeLists.txt', '.yujin_init_build']]:
            if os.path.isfile(f):
                os.remove(f)
        for d in [os.path.join(dir_to_be_cleaned, x) for x in ['build', 'devel']]:
            if os.path.isdir(d):
                shutil.rmtree(d)
        console.pretty_println('done.\n', console.yellow)
    else:
        if os.path.isdir(os.path.abspath(dir_to_be_cleaned)):
            console.pretty_print("Cleaning build directory : ", console.cyan)
            console.pretty_println(dir_to_be_cleaned, console.yellow)
            shutil.rmtree(os.path.abspath(dir_to_be_cleaned))
        else:
            console.logerror("Could not clean %s [does not exist]" % dir_to_be_cleaned)


def init_build():
    args = parse_arguments()

    ##########################
    # Default underlays
    ##########################
    if args.no_default_underlay:
        args.default_underlay = None
    elif not args.default_underlay:
        args.default_underlay = settings.get_default_track()

    ##########################
    # Toolchains and Platform
    ##########################
    if args.list_toolchains:
        list_toolchains()
        return
    if args.list_platforms:
        list_platforms()
        return
    if args.clean:
        clean(args.dir, args.sources)
        return
    init_configured_build(args.default_underlay, args.dir, args.sources, args.underlays, args.install, args.doc, args.release, args.toolchain, args.platform)



##############################################################################
# Imports
##############################################################################

import argparse
import sys

from . import ansible_pc

##############################################################################
# Methods
##############################################################################


def main(args=sys.argv[1:]):
    # Manually handle exclusive options first since subparsers can't be 'optional'.
    if len(sys.argv) == 2 and sys.argv[1] == "--version":
        print("Version: TODO")
        return
    if len(sys.argv) == 2 and sys.argv[1] == "--update":
        print("Update: TODO")
        return

    # TODO check that yujin_ansible_bootstrap has already been executed, and if not - execute and exit with a friendly message
    parser = argparse.ArgumentParser(description='Executor for ansible playbooks in a wild variety of situations.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)  # RawTextHelpFormatter for whitespace preservation

    # These aren't actually used, but we include them here so they get onto the help screen
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-v', '--version', action='store_true', help='Display the version of the installed yujin ansible playbooks package')
    group.add_argument('-u', '--update', action='store_true', help='Make sure you have the latest playbooks')

    subparsers = parser.add_subparsers(title='playbooks',
                                       help='name of an existing playbook',
                                       metavar="<playbook>")
    ansible_pc.add_subparser(subparsers)
    options, unused_unknown_args = parser.parse_known_args(args)
    options.func(options)  # relay arg parsing to the subparser configured `set_defaults` function callback



##############################################################################
# Imports
##############################################################################

#import os
import os.path

import re
import sys
#import stat  # file permissions
import argparse
#from argparse import RawTextHelpFormatter
#import tempfile
import shutil
import multiprocessing
import subprocess
from catkin_pkg.packages import find_packages

##############################################################################
# Local imports
##############################################################################

import console
import config_cache
import common
import settings
from make_doc import make_doc

import catkin_make.terminal_color as terminal_color
from catkin_make.terminal_color import fmt
import catkin_make.builder as builder  # extract_cmake_and_make_arguments, cmake_input_changed

##############################################################################
# Methods
##############################################################################


def _parse_args(args=sys.argv[1:]):
    args, cmake_args, make_args = builder.extract_cmake_and_make_arguments(args)

    parser = argparse.ArgumentParser(description='Creates the catkin workspace layout and invokes cmake and make. Any argument starting with "-D" will be passed to the "cmake" invocation. All other arguments (i.e. target names) are passed to the "make" invocation.')
    parser.add_argument('-j', '--jobs', type=int, metavar='JOBS', default=None, nargs='?', help='Specifies the number of jobs (commands) to run simultaneously. Defaults to the environment variable ROS_PARALLEL_JOBS and falls back to the number of CPU cores.')
    parser.add_argument('--force-cmake', action='store_true', help='Invoke "cmake" even if it has been executed before [false]')
    parser.add_argument('-p', '--pre-clean', action='store_true', help='Clean build temporaries before making [false]')
    parser.add_argument('-c', '--cmake-only', action='store_true', help='Do not compile, just force a re-run of cmake [false]')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-i', '--install', action='store_true', help='Run install step after making [false]')
    group.add_argument('--track', choices=settings.VALID_TRACKS, dest='default_underlay', action='store', default=None, help='convenience equivalent for the --default-underlay option')
    group.add_argument('--install-rosdeps-track', choices=settings.VALID_TRACKS, action='store', default=None, help='Install all rosdeps for the workspace sources and given track [None]')
    group.add_argument('--install-rosdeps', action='store_true', help='Install all rosdeps for the workspace sources and track set by `yujin_tools_settings --get-default-track` [false]')
    group.add_argument('-t', '--tests', action='store_true', help='Make tests [false]')
    group.add_argument('-r', '--run-tests', action='store_true', help='Run tests (does not build them) [false]')
    parser.add_argument('--strip', action='store_true', help='Strips binaries, only valid with --install')
    parser.add_argument('--no-color', action='store_true', help='Disables colored ouput')
    parser.add_argument('--target', default=None, help='Build against a particular target only')
    parser.add_argument('--pkg', help='Invoke "make" on a specific package only')

    docgroup = parser.add_mutually_exclusive_group()
    docgroup.add_argument('-d', '--doc', action='store_true', help='Generates the documents for packages in the workspace.')
    docgroup.add_argument('--doc-only', action='store_true', help='Generates the documents for packages in the workspace. Does not build source')

    parser.add_argument('--cmake-args', dest='cmake_args', nargs='*', type=str,
        help='Arbitrary arguments which are passes to CMake. It must be passed after other arguments since it collects all following options.')
    parser.add_argument('--make-args', dest='make_args', nargs='*', type=str,
        help='Arbitrary arguments which are passes to make. It must be passed after other arguments since it collects all following options. This is only necessary in combination with --cmake-args since else all unknown arguments are passed to make anyway.')

    options, unknown_args = parser.parse_known_args(args)
    if not options.jobs and not [a for a in args if a == '-j' or a == '--jobs']:
        options.jobs = common.good_number_of_jobs()
    options.cmake_args = cmake_args
    options.make_args = unknown_args + make_args
    return options


def validate_build_space(base_path):
    if not os.path.isfile(os.path.join(base_path, 'config.cmake')):
        raise RuntimeError('Switch to a valid build directory (must contain a config.cmake cmake cache file).')

    # verify that the base path does not contain a CMakeLists.txt
    if os.path.exists(os.path.join(base_path, 'CMakeLists.txt')):
        raise RuntimeError('Switch to a valid build directory (this one is a cmake project, i.e. contains a CMakeLists.txt).')

    # verify that the base path does not contain a package.xml
    if os.path.exists(os.path.join(base_path, 'package.xml')):
        raise RuntimeError('Switch to a valid build directory (this one is a catkin package, i.e. contains a package.xml).')

    # this will have been generated already by yujin_init_build
    source_path = os.path.join(base_path, 'src')
    if os.path.exists(source_path):
        if os.path.islink(source_path):
            absolute_source_path = os.readlink(source_path)
        else:
            absolute_source_path = source_path
        if not os.path.exists(absolute_source_path):
            raise RuntimeError('The specified source space does not exist [%s]' % absolute_source_path)
    else:
        raise RuntimeError('Could not find a valid source path (did you init build correctly?)')


def check_and_update_source_repo_paths(build_source_path):
    '''
      Check that the symbolic links we possible created in the build src directory still
      match the original source directory. If not, update.
    '''
    f = open(os.path.join(build_source_path, '.yujin_init_build'), 'r')
    try:
        rel_path = f.read()
    finally:
        f.close()
    original_source_path = os.path.abspath(os.path.join(build_source_path, rel_path))
    # broken links show up as files
    (unused_root, build_source_subdirectories, files) = os.walk(build_source_path).next()
    for f in files:
        if common.is_broken_symlink(os.path.join(build_source_path, f)):
            build_source_subdirectories.append(f)
    original_source_subdirectories = os.walk(original_source_path).next()[1]
    #print build_source_subdirectories
    #print original_source_subdirectories
    removed = [d for d in build_source_subdirectories if d not in original_source_subdirectories]
    added = [d for d in original_source_subdirectories if d not in build_source_subdirectories]
    for d in removed:
        os.unlink(os.path.join(build_source_path, d))
    for d in added:
        common.create_symlink(os.path.join(original_source_path, d), os.path.join(build_source_path, d), quiet=True)


def insert_yujin_make_signature(yujin_make_root, devel_path):
    '''
      Insert YUJIN_MAKE_ROOT=xxx into devel/setup.bash if the cmake process succeeded. This helps yujin_make
      kickstart itself from anywhere in the future.
    '''
    setup_sh = open(os.path.join(devel_path, 'setup.sh'), 'a+')
    found = False
    for line in setup_sh:
        if re.search('^export YUJIN_MAKE_ROOT', line):
            found = True
            break
    if not found:
        setup_sh.write("export YUJIN_MAKE_ROOT=%s\n" % yujin_make_root)


def install_rosdeps(base_path, source_path, rosdistro, no_color):
    # -r continue even with errors
    cmd = ['rosdep', 'install']  # if you want it to continue installing despite errors, '-r']
    underlays = config_cache.get_underlays_list_from_config_cmake(base_path)
    print("Rosdep Search Paths")
    for underlay in underlays:
        underlay_path = underlay
        underlay_source_path = os.path.abspath(os.path.join(underlay, os.pardir, "src"))
        if os.path.isdir(underlay_source_path):
	    underlay_path = underlay_source_path
        if os.path.isdir(underlay_path):
            cmd += ['--from-paths', underlay_path]
	    print(" - adding underlay '%s'" % underlay_path)
        else:
	    print(" - not adding underlay '%s' [not found]" % underlay_path)
    cmd += ['--from-paths', source_path, '--ignore-src', '--rosdistro', rosdistro, '-y']
    env = os.environ.copy()
    try:
        builder.print_command_banner(cmd, source_path, color=not no_color)
        if no_color:
            builder.run_command(cmd, source_path, env=env)
        else:
            builder.run_command_colorized(cmd, source_path, env=env)
    except subprocess.CalledProcessError:
        return 1  # rosdep will already have provided its own error message, no need to regurgitate
    return 0


def make_main():
    args = _parse_args()
    cmake_args = args.cmake_args

    if args.no_color:
        terminal_color.disable_ANSI_colors()

    (base_path, build_path, devel_path, source_path) = common.get_default_paths()
    doc_path = config_cache.get_doc_prefix_from_config_cmake(base_path)

    validate_build_space(base_path)  # raises a RuntimeError if there is a problem

    # Install rosdeps if requested
    if args.install_rosdeps:
        return install_rosdeps(base_path, source_path, settings.get_default_track(), args.no_color)
    if args.install_rosdeps_track is not None:
        return install_rosdeps(source_path, args.install_rosdeps_track, args.no_color)


    # Clear out previous temporaries if requested
    if args.pre_clean:
        console.pretty_print("Pre-cleaning before building.", console.cyan)
        shutil.rmtree(devel_path, ignore_errors=True)
        shutil.rmtree(build_path, ignore_errors=True)
        shutil.rmtree(doc_path, ignore_errors=True)

    # check for new build
    if not os.path.exists(build_path):
        os.mkdir(build_path)
    #if not os.path.exists(devel_path):
    #    os.mkdir(devel_path)

    # ensure toplevel cmake file exists
    toplevel_cmake = os.path.join(source_path, 'CMakeLists.txt')
    if not os.path.exists(toplevel_cmake):
        return fmt('@{rf}No toplevel cmake file@')

    # did source paths get added to the original location?
    check_and_update_source_repo_paths(source_path)

    packages = find_packages(source_path, exclude_subspaces=True)

    # verify that specified package exists in workspace
    if args.pkg:
        packages_by_name = {p.name: path for path, p in packages.iteritems()}
        if args.pkg not in packages_by_name:
            raise RuntimeError('Package %s not found in the workspace' % args.pkg)

    # check if cmake must be run (either for a changed list of package paths or changed cmake arguments)
    force_cmake, _ = builder.cmake_input_changed(packages, build_path, cmake_args=cmake_args)

    # check if toolchain.cmake, config.cmake exist
    toolchain_cmd = "-DCMAKE_TOOLCHAIN_FILE=%s" % os.path.join(base_path, 'toolchain.cmake') if os.path.isfile(os.path.join(base_path, 'toolchain.cmake')) else None
    config_cmd = "-C%s" % os.path.join(base_path, 'config.cmake') if os.path.isfile(os.path.join(base_path, 'config.cmake')) else None

    # Help find catkin cmake and python
    unused_catkin_toplevel, catkin_python_path, unused_catkin_cmake_path = common.find_catkin(base_path)
    pkg_config_paths = common.generate_pkg_config_path(base_path)
    env = os.environ.copy()
    # PYTHONPATH
    # Don't add to the environment variable - this mucks up catkin's catkin_generated/setup_cached.py
    # environment later (how? I can't remember - something to do with the default underlay).
    # Maybe we can do away with this now catkin can look up install spaces?
    #try:
    #    env['PYTHONPATH'] = env['PYTHONPATH'] + os.pathsep + catkin_python_path
    #except KeyError:
    #    env['PYTHONPATH'] = catkin_python_path
    sys.path.append(catkin_python_path)
    # PKG_CONFIG_PATH
    for path in pkg_config_paths:
        try:
            env['PKG_CONFIG_PATH'] = env['PKG_CONFIG_PATH'] + os.pathsep + path
        except KeyError:
            env['PKG_CONFIG_PATH'] = path

    if args.doc_only:
        console.pretty_println('Generates documents only', console.bold_white)
        make_doc(source_path, doc_path, packages)
        return

    # consider calling cmake
    makefile = os.path.join(build_path, 'Makefile')
    if not os.path.exists(makefile) or args.force_cmake or force_cmake:
        cmd = ['cmake', source_path]
        if toolchain_cmd:
            cmd.append(toolchain_cmd)
        if config_cmd:
            cmd.append(config_cmd)
        cmd += cmake_args

        #new_env = common.generate_underlays_environment(base_path)
        try:
            builder.print_command_banner(cmd, build_path, color=not args.no_color)
            if args.no_color:
                builder.run_command(cmd, build_path, env=env)
            else:
                builder.run_command_colorized(cmd, build_path, env=env)
        except subprocess.CalledProcessError:
            return fmt('@{rf}Invoking @{boldon}"cmake"@{boldoff} failed')
    else:
        cmd = ['make', 'cmake_check_build_system']
        #new_env = common.generate_environment(base_path) # underlays + current workspace
        try:
            builder.print_command_banner(cmd, build_path, color=not args.no_color)
            if args.no_color:
                builder.run_command(cmd, build_path, env=env)
            else:
                builder.run_command_colorized(cmd, build_path, env=env)
        except subprocess.CalledProcessError:
            return fmt('@{rf}Invoking @{boldon}"make cmake_check_build_system"@{boldoff} failed')

    insert_yujin_make_signature(base_path, devel_path)

    # invoke make
    if not args.cmake_only:
        if args.target:
            cmd = ['make', args.target]
        elif args.install:
            cmd = ['make', 'install/strip'] if args.strip else ['make', 'install']
        elif args.tests:
            cmd = ['make', 'tests']
        elif args.run_tests:
            cmd = ['make', 'test']
        else:
            cmd = ['make']
        jobs = args.jobs
        if args.jobs == '':
            cmd.append('-j')
        else:
            jobs = args.jobs
            if not jobs:
                if 'ROS_PARALLEL_JOBS' in os.environ:
                    ros_parallel_jobs = os.environ['ROS_PARALLEL_JOBS']
                    cmd += [arg for arg in ros_parallel_jobs.split(' ') if arg]
                else:
                    jobs = multiprocessing.cpu_count()
            if jobs:
                cmd.append('-j%d' % jobs)
                cmd.append('-l%d' % jobs)
        cmd += args.make_args
        try:
            make_path = build_path
            if args.pkg:
                make_path = os.path.join(make_path, packages_by_name[args.pkg])
            builder.print_command_banner(cmd, make_path, color=not args.no_color)
            builder.run_command(cmd, make_path, env=env)
        except subprocess.CalledProcessError:
            return fmt('@{rf}Invoking @{boldon}"make"@{boldoff} failed')

    if args.doc:
        make_doc(source_path, doc_path, packages)
